/*
 * Wire
 * Copyright (C) 2024 Wire Swiss GmbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 */
package com.wire.android.ui.home.conversations.call

import androidx.lifecycle.SavedStateHandle
import com.wire.android.config.CoroutineTestExtension
import com.wire.android.config.NavigationTestExtension
import com.wire.android.ui.home.conversations.ConversationNavArgs
import com.wire.android.ui.home.conversations.details.participants.usecase.ObserveParticipantsForConversationUseCase
import com.wire.android.ui.navArgs
import com.wire.kalium.logic.data.id.ConversationId
import com.wire.kalium.logic.feature.call.usecase.AnswerCallUseCase
import com.wire.kalium.logic.feature.call.usecase.EndCallUseCase
import com.wire.kalium.logic.feature.call.usecase.IsEligibleToStartCallUseCase
import com.wire.kalium.logic.feature.call.usecase.ObserveEstablishedCallsUseCase
import com.wire.kalium.logic.feature.call.usecase.ObserveOngoingCallsUseCase
import com.wire.kalium.logic.feature.conversation.ObserveConversationDetailsUseCase
import com.wire.kalium.logic.feature.conversation.ObserveDegradedConversationNotifiedUseCase
import com.wire.kalium.logic.feature.conversation.SetUserInformedAboutVerificationUseCase
import com.wire.kalium.logic.sync.ObserveSyncStateUseCase
import io.mockk.MockKAnnotations
import io.mockk.coEvery
import io.mockk.coVerify
import io.mockk.every
import io.mockk.impl.annotations.MockK
import kotlinx.coroutines.flow.emptyFlow
import kotlinx.coroutines.flow.flowOf
import org.amshove.kluent.internal.assertEquals
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.extension.ExtendWith

@ExtendWith(CoroutineTestExtension::class)
@ExtendWith(NavigationTestExtension::class)
class ConversationListCallViewModelTest {

    @MockK
    private lateinit var savedStateHandle: SavedStateHandle

    @MockK
    private lateinit var observeOngoingCalls: ObserveOngoingCallsUseCase

    @MockK
    private lateinit var observeEstablishedCalls: ObserveEstablishedCallsUseCase

    @MockK
    private lateinit var joinCall: AnswerCallUseCase

    @MockK
    private lateinit var endCall: EndCallUseCase

    @MockK
    private lateinit var observeSyncState: ObserveSyncStateUseCase

    @MockK
    private lateinit var isConferenceCallingEnabled: IsEligibleToStartCallUseCase

    @MockK(relaxed = true)
    private lateinit var onAnswered: (conversationId: ConversationId) -> Unit

    @MockK
    private lateinit var observeConversationDetails: ObserveConversationDetailsUseCase

    @MockK
    private lateinit var observeParticipantsForConversation: ObserveParticipantsForConversationUseCase

    @MockK
    lateinit var setUserInformedAboutVerificationUseCase: SetUserInformedAboutVerificationUseCase

    @MockK
    lateinit var observeDegradedConversationNotifiedUseCase: ObserveDegradedConversationNotifiedUseCase

    private lateinit var conversationListCallViewModel: ConversationListCallViewModel

    @BeforeEach
    fun setUp() {
        MockKAnnotations.init(this)
        val conversationId = ConversationId("some-dummy-value", "some.dummy.domain")
        every { savedStateHandle.navArgs<ConversationNavArgs>() } returns ConversationNavArgs(conversationId = conversationId)
        coEvery { observeEstablishedCalls.invoke() } returns emptyFlow()
        coEvery { observeOngoingCalls.invoke() } returns emptyFlow()
        coEvery { observeConversationDetails(any()) } returns flowOf()
        coEvery { observeParticipantsForConversation(any()) } returns flowOf()
        coEvery { setUserInformedAboutVerificationUseCase(any()) } returns Unit
        coEvery { observeDegradedConversationNotifiedUseCase(any()) } returns flowOf(false)

        conversationListCallViewModel = ConversationListCallViewModel(
            savedStateHandle = savedStateHandle,
            observeOngoingCalls = observeOngoingCalls,
            observeEstablishedCalls = observeEstablishedCalls,
            answerCall = joinCall,
            endCall = endCall,
            observeSyncState = observeSyncState,
            isConferenceCallingEnabled = isConferenceCallingEnabled,
            observeConversationDetails = observeConversationDetails,
            observeParticipantsForConversation = observeParticipantsForConversation,
            setUserInformedAboutVerification = setUserInformedAboutVerificationUseCase,
            observeDegradedConversationNotified = observeDegradedConversationNotifiedUseCase
        )
    }

    @Test
    fun `given join dialog displayed, when user dismiss it, then hide it`() {
        conversationListCallViewModel.conversationCallViewState = conversationListCallViewModel.conversationCallViewState.copy(
            shouldShowJoinAnywayDialog = true
        )

        conversationListCallViewModel.dismissJoinCallAnywayDialog()

        assertEquals(false, conversationListCallViewModel.conversationCallViewState.shouldShowJoinAnywayDialog)
    }

    @Test
    fun `given no ongoing call, when user tries to join a call, then invoke answerCall call use case`() {
        conversationListCallViewModel.conversationCallViewState =
            conversationListCallViewModel.conversationCallViewState.copy(hasEstablishedCall = false)

        coEvery { joinCall(conversationId = any()) } returns Unit

        conversationListCallViewModel.joinOngoingCall(onAnswered)

        coVerify(exactly = 1) { joinCall(conversationId = any()) }
        coVerify(exactly = 1) { onAnswered(any()) }
        assertEquals(false, conversationListCallViewModel.conversationCallViewState.shouldShowJoinAnywayDialog)
    }

    @Test
    fun `given an ongoing call, when user tries to join a call, then show JoinCallAnywayDialog`() {
        conversationListCallViewModel.conversationCallViewState =
            conversationListCallViewModel.conversationCallViewState.copy(hasEstablishedCall = true)

        conversationListCallViewModel.joinOngoingCall(onAnswered)

        assertEquals(true, conversationListCallViewModel.conversationCallViewState.shouldShowJoinAnywayDialog)
        coVerify(inverse = true) { joinCall(conversationId = any()) }
    }

    @Test
    fun `given an ongoing call, when user confirms dialog to join a call, then end current call and join the newer one`() {
        conversationListCallViewModel.conversationCallViewState =
            conversationListCallViewModel.conversationCallViewState.copy(hasEstablishedCall = true)
        conversationListCallViewModel.establishedCallConversationId = ConversationId("value", "Domain")
        coEvery { endCall(any()) } returns Unit

        conversationListCallViewModel.joinAnyway(onAnswered)

        coVerify(exactly = 1) { endCall(any()) }
    }
}
